# Generated by Django 2.2.19 on 2021-03-29 12:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eri", "0002_auto_20210322_0233"),
    ]

    operations = [
        migrations.AddField(
            model_name="logentry",
            name="version",
            field=models.CharField(default="", max_length=33, verbose_name="状态版本"),
        ),
    ]
