# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2020-02-13 07:43
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("component_framework", "0005_auto_20190723_1806"),
    ]

    operations = [
        migrations.AlterField(
            model_name="componentmodel",
            name="version",
            field=models.CharField(default="legacy", max_length=64, verbose_name="组件版本"),
        ),
    ]
