# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-12-01 07:06
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("statistics", "0012_auto_20201123_1552"),
    ]

    operations = [
        migrations.AlterField(
            model_name="componentexecutedata",
            name="component_code",
            field=models.CharField(db_index=True, max_length=255, verbose_name="组件编码"),
        ),
        migrations.AlterField(
            model_name="componentexecutedata",
            name="instance_id",
            field=models.CharField(db_index=True, max_length=32, verbose_name="实例ID"),
        ),
    ]
